"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.unpinDevDeps = void 0;
const getWorkspaceInfo_1 = require("../utils/getWorkspaceInfo");
const partialClonePackageInfo_1 = require("../utils/partialClonePackageInfo");
const writePackageJsonUpdates_1 = require("../utils/writePackageJsonUpdates");
const collectDevDeps_1 = require("../utils/collectDevDeps");
/**
 * Get a copy of `packageInfo` with any matching dev deps updated.
 * @param packageInfo package.json for an individual package
 * @param updateDevDeps dev dep versions to change
 * @returns new package info if there were any updates, or undefined if not
 */
function getUpdatedPackageInfo(packageInfo, updateDevDeps) {
    if (!packageInfo.devDependencies) {
        return;
    }
    let updatedInfo;
    for (const { name, oldVersion, newVersion } of updateDevDeps) {
        if (packageInfo.devDependencies[name] === oldVersion) {
            updatedInfo ?? (updatedInfo = (0, partialClonePackageInfo_1.partialClonePackageInfo)(packageInfo, ['devDependencies']));
            updatedInfo.devDependencies[name] = newVersion;
        }
    }
    return updatedInfo;
}
function unpinDevDeps(options) {
    const { exclude = [], patch = [], write = true } = options;
    const workspaceInfo = (0, getWorkspaceInfo_1.getWorkspaceInfo)();
    const { rootPackageInfo, packageInfos } = workspaceInfo;
    // collect external dev deps from the root package.json and individual packages
    const externalDevDeps = (0, collectDevDeps_1.collectDevDeps)(workspaceInfo, exclude);
    // collect versions to update (use an array in case of multiple versions of a dep)
    const updateDevDeps = [];
    for (const [name, versions] of Object.entries(externalDevDeps)) {
        for (const oldVersion of Object.keys(versions)) {
            // update if it's an exact version and not prerelease
            if (/^\d+\.\d+\.\d+$/.test(oldVersion)) {
                const newVersion = patch.includes(name) ? `~${oldVersion}` : `^${oldVersion}`;
                console.log(`Updating ${name}@${oldVersion} to ${newVersion}`);
                updateDevDeps.push({ name, oldVersion, newVersion });
            }
        }
    }
    // update individual packages
    const updatedPackageInfos = [rootPackageInfo, ...Object.values(packageInfos)]
        .map((packageInfo) => getUpdatedPackageInfo(packageInfo, updateDevDeps))
        .filter((p) => !!p);
    if (write) {
        (0, writePackageJsonUpdates_1.writePackageJsonUpdates)(updatedPackageInfos);
    }
    return updatedPackageInfos;
}
exports.unpinDevDeps = unpinDevDeps;
